from maggma.api.query_operator.dynamic import NumericQuery
from maggma.api.resource import ReadOnlyResource
from emmet.core.thermo import ThermoDoc

from maggma.api.query_operator import (
    PaginationQuery,
    SortQuery,
    SparseFieldsQuery,
)
from mp_api.routes.thermo.query_operators import IsStableQuery

from mp_api.routes.materials.query_operators import MultiMaterialIDQuery, FormulaQuery


def thermo_resource(thermo_store):
    resource = ReadOnlyResource(
        thermo_store,
        ThermoDoc,
        query_operators=[
            MultiMaterialIDQuery(),
            FormulaQuery(),
            IsStableQuery(),
            NumericQuery(model=ThermoDoc),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(
                ThermoDoc, default_fields=["material_id", "last_updated"]
            ),
        ],
        tags=["Thermo"],
    )

    return resource



from maggma.api.resource import ReadOnlyResource
from mp_api.routes.wulff.models import WulffDoc

from maggma.api.query_operator import PaginationQuery, SparseFieldsQuery


def wulff_resource(wulff_store):
    resource = ReadOnlyResource(
        wulff_store,
        WulffDoc,
        query_operators=[
            PaginationQuery(),
            SparseFieldsQuery(WulffDoc, default_fields=["task_id"]),
        ],
        tags=["Surface Properties"],
        enable_default_search=False,
    )

    return resource



import pytest
from mp_api.core.client import BaseRester
from mp_api.matproj import MPRester


@pytest.fixture
def rester():
    rester = BaseRester()
    yield rester
    rester.session.close()


@pytest.fixture()
def mpr():
    rester = MPRester()
    yield rester
    rester.session.close()


@pytest.mark.xfail
def test_post_fail(rester):
    rester._post_resource({}, suburl="materials/find_structure")


def test_pagination(mpr):
    mpids = mpr.materials.search_material_docs(
        all_fields=False, fields=["material_id"], num_chunks=2, chunk_size=1000
    )
    assert len(mpids) > 1000


def test_count(mpr):
    count = mpr.materials.count(
        dict(task_ids="mp-149", all_fields=False, fields="material_id", limit=1000)
    )
    assert count == 1


@pytest.mark.xfail
def test_get_document_no_id(mpr):
    mpr.materials.get_document_by_id(None)


@pytest.mark.xfail
def test_get_document_no_doc(mpr):
    mpr.materials.get_document_by_id("mp-1a")


def test_available_fields(rester, mpr):
    assert len(mpr.materials.available_fields) > 0
    assert rester.available_fields == ["Unknown fields."]



from typing import List, Optional
from emmet.core.xas import XASDoc


class XASRester:

    def get_document_by_id(
            self,
            document_id: str,
            fields: Optional[List[str]] = None,
            monty_decode: bool = True,
    ) -> XASDoc:
        ...



from mp_api.core.client import BaseRester, MPRestError
from mp_api.routes.synthesis.models import SynthesisSearchResultModel
from typing import List


class SynthesisRester(BaseRester):

    suffix = "synthesis"
    document_model = SynthesisSearchResultModel  # type: ignore

    def search_synthesis_text(self, keywords: List[str]):
        """
        Search synthesis recipe text.
        Arguments:
            keywords (List[str]): List of search keywords
        Returns:
            synthesis_docs ([SynthesisDoc]): List of synthesis documents
        """

        keyword_string = ",".join(keywords)

        synthesis_docs = self._query_resource(
            criteria={"keywords": keyword_string}, use_document_model=True,
        ).get("data", None)

        if synthesis_docs is None:
            raise MPRestError("Cannot find any matches.")

        return synthesis_docs



from mp_api.core.client import BaseRester
from mp_api.routes.phonon.models import PhononBSDoc, PhononImgDoc


class PhononRester(BaseRester):

    suffix = "phonon"
    document_model = PhononBSDoc  # type: ignore
    primary_key = "task_id"


class PhononImgRester(BaseRester):

    suffix = "phonon_img"
    document_model = PhononImgDoc  # type: ignore
    primary_key = "task_id"



from typing import List
from datetime import datetime
from monty.json import MontyDecoder

from pydantic import BaseModel, Field, validator
from pymatgen.core import Structure
from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine


class PhononBS(BaseModel):
    """
    Model for a phonon band structure object
    """

    eigendisplacements: dict = Field(
        None, description="Phonon eigendisplacements in cartesian coordinates",
    )

    has_nac: bool = Field(
        None, description="Whether non-analytical corrections at Gamma are included",
    )

    bands: List[List[float]] = Field(
        None, description="Phonon band eigenvalues in eV",
    )

    qpoints: List[List[float]] = Field(
        None,
        description="List of q-points in fractional coordinates of the reciprocal lattice",
    )

    labels_dict: dict = Field(
        None, description="q-point labels dictionary",
    )

    lattice_rec: dict = Field(
        None, description="Reciprocal lattice of the structure",
    )

    structure: Structure = Field(
        None, description="Structure of the material",
    )

    class Config:
        extra = "allow"


class PhononBSDoc(BaseModel):
    """
    Phonon band structures.
    """

    task_id: str = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******",
    )

    ph_bs: PhononBandStructureSymmLine = Field(
        None, description="Phonon band structure object",
    )

    last_updated: datetime = Field(
        None,
        description="Timestamp for the most recent calculation for this Material document",
    )

    # Make sure that the datetime field is properly formatted
    @validator("last_updated", pre=True)
    def last_updated_dict_ok(cls, v):
        return MontyDecoder().process_decoded(v)


class PhononImgDoc(BaseModel):
    """
    Model for a document containing phonon image data.
    """

    plot: bytes = Field(
        None, description="Plot image data.",
    )

    task_id: str = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******",
    )

    last_updated: datetime = Field(
        None, description="Timestamp for the most recent calculation for this document",
    )

    # Make sure that the datetime field is properly formatted
    @validator("last_updated", pre=True)
    def last_updated_dict_ok(cls, v):
        return MontyDecoder().process_decoded(v)

    # Make sure that the plot field is properly formatted
    @validator("plot", pre=True)
    def plot_bytes_ok(cls, v):
        return str(v)



from typing import List, Optional
from emmet.core.thermo import ThermoDoc


class ThermoRester:

    def get_document_by_id(
        self, document_id: str, fields: Optional[List[str]] = None, monty_decode: bool = True,
    ) -> ThermoDoc: ...



from ratelimit import sleep_and_retry, limits
from mp_api.core.settings import MAPISettings


@sleep_and_retry
@limits(calls=MAPISettings().requests_per_min, period=60)
def check_limit():
    """
    Empty function for enabling global rate limiting.
    """
    return



from maggma.api.resource import ReadOnlyResource
from mp_api.routes.dielectric.models import DielectricDoc

from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery
from mp_api.routes.dielectric.query_operators import DielectricQuery


def dielectric_resource(dielectric_store):
    resource = ReadOnlyResource(
        dielectric_store,
        DielectricDoc,
        query_operators=[
            DielectricQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(DielectricDoc, default_fields=["task_id", "last_updated"]),
        ],
        tags=["Dielectric"],
    )

    return resource



from mp_api.core.client import BaseRester
from mp_api.routes.wulff.models import WulffDoc


class WulffRester(BaseRester):

    suffix = "wulff"
    document_model = WulffDoc  # type: ignore
    primary_key = "task_id"



from typing import List, Optional
from mp_api.routes.magnetism.models import MagnetismDoc


class MagnetismRester:

    def query_by_task_id(
        self, task_id, fields: Optional[List[str]] = None, monty_decode: bool = True
    ) -> MagnetismDoc: ...



from typing import List
from datetime import datetime
from monty.json import MontyDecoder

from pydantic import BaseModel, Field, validator


class PiezoData(BaseModel):
    """
    Piezoelectric tensor and associated information.
    """

    total: List[List[float]] = Field(
        None,
        description="Total piezoelectric tensor in C/m²",
    )

    ionic: List[List[float]] = Field(
        None,
        description="Ionic contribution to piezoelectric tensor in C/m²",
    )

    static: List[List[float]] = Field(
        None,
        description="Electronic contribution to piezoelectric tensor in C/m²",
    )

    e_ij_max: float = Field(
        None,
        description="Piezoelectric modulus",
    )

    max_direction: List[float] = Field(
        None,
        description="Crystallographic direction",
    )


class PiezoDoc(BaseModel):
    """
    Model for a document containing piezoelectric data
    """

    piezo: PiezoData = Field(
        None,
        description="Piezoelectric data",
    )

    task_id: str = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******",
    )

    last_updated: datetime = Field(
        None,
        description="Timestamp for the most recent calculation for this Material document",
    )

    # Make sure that the datetime field is properly formatted
    @validator("last_updated", pre=True)
    def last_updated_dict_ok(cls, v):
        return MontyDecoder().process_decoded(v)



from os import environ
import warnings
from typing import Optional, Tuple, List
from enum import Enum, unique

from pymatgen.core import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core.surface import get_symmetrically_equivalent_miller_indices
from pymatgen.analysis.magnetism import Ordering
from emmet.core.mpid import MPID
from emmet.core.symmetry import CrystalSystem

from mp_api.core.client import BaseRester
from mp_api.routes import *

_DEPRECATION_WARNING = (
    "MPRester is being modernized. Please use the new method suggested and "
    "read more about these changes at https://docs.materialsproject.org/api. The current "
    "methods will be retained until at least January 2022 for backwards compatibility."
)


DEFAULT_API_KEY = environ.get("MP_API_KEY", None)
DEFAULT_ENDPOINT = environ.get("MP_API_ENDPOINT", "https://api.materialsproject.org/")


@unique
class TaskType(Enum):
    """task types available in MP"""

    GGA_OPT = "GGA Structure Optimization"
    GGAU_OPT = "GGA+U Structure Optimization"
    SCAN_OPT = "SCAN Structure Optimization"
    GGA_LINE = "GGA NSCF Line"
    GGAU_LINE = "GGA+U NSCF Line"
    GGA_UNIFORM = "GGA NSCF Uniform"
    GGAU_UNIFORM = "GGA+U NSCF Uniform"
    GGA_STATIC = "GGA Static"
    GGAU_STATIC = "GGA+U Static"
    GGA_STATIC_DIEL = "GGA Static Dielectric"
    GGAU_STATIC_DIEL = "GGA+U Static Dielectric"
    GGA_DEF = "GGA Deformation"
    GGAU_DEF = "GGA+U Deformation"
    LDA_STATIC_DIEL = "LDA Static Dielectric"


class MPRester:
    """
    Intended as a drop-in replacement for the current MPRester.
    """

    def __init__(
        self,
        api_key=DEFAULT_API_KEY,
        endpoint=DEFAULT_ENDPOINT,
        notify_db_version=True,
        include_user_agent=True,
    ):
        """
        Args:
            api_key (str): A String API key for accessing the MaterialsProject
                REST interface. Please obtain your API key at
                https://next-gen.materialsproject.org/api. If this is None,
                the code will check if there is a "MP_API_KEY" setting.
                If so, it will use that environment variable. This makes
                easier for heavy users to simply add this environment variable to
                their setups and MPRester can then be called without any arguments.
            endpoint (str): Url of endpoint to access the MaterialsProject REST
                interface. Defaults to the standard Materials Project REST
                address at "https://api.materialsproject.org", but
                can be changed to other urls implementing a similar interface.
            notify_db_version (bool): If True, the current MP database version will
                be retrieved and logged locally in the ~/.pmgrc.yaml. If the database
                version changes, you will be notified. The current database version is
                also printed on instantiation. These local logs are not sent to
                materialsproject.org and are not associated with your API key, so be
                aware that a notification may not be presented if you run MPRester
                from multiple computing environments.
            include_user_agent (bool): If True, will include a user agent with the
                HTTP request including information on pymatgen and system version
                making the API request. This helps MP support pymatgen users, and
                is similar to what most web browsers send with each page request.
                Set to False to disable the user agent.
        """

        self.api_key = api_key
        self.endpoint = endpoint
        self.session = BaseRester._create_session(
            api_key=api_key, include_user_agent=include_user_agent
        )

        self._all_resters = []

        for cls in BaseRester.__subclasses__():

            rester = cls(
                api_key=api_key,
                endpoint=endpoint,
                include_user_agent=include_user_agent,
                session=self.session,
            )

            self._all_resters.append(rester)

            setattr(
                self, cls.suffix.replace("/", "_"), rester,
            )

        self.find_structure = self.materials.find_structure

        self.get_bandstructure_by_material_id = (
            self.electronic_structure_bandstructure.get_bandstructure_from_material_id
        )
        self.get_dos_by_material_id = (
            self.electronic_structure_dos.get_dos_from_material_id
        )

        self.get_charge_density_from_calculation_id = (
            self.charge_density.get_charge_density_from_calculation_id
        )

        self.get_charge_density_calculation_details = (
            self.charge_density.get_charge_density_calculation_details
        )

        self.get_charge_density_calculation_ids_from_material_id = (
            self.charge_density.get_charge_density_calculation_ids_from_material_id
        )

    def __enter__(self):
        """
        Support for "with" context.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Support for "with" context.
        """
        self.session.close()

    def get_structure_by_material_id(
        self, material_id, final=True, conventional_unit_cell=False
    ) -> Structure:
        """
        Get a Structure corresponding to a material_id.

        Args:
            material_id (str): Materials Project material_id (a string,
                e.g., mp-1234).
            final (bool): Whether to get the final structure, or the list of initial
                (pre-relaxation) structures. Defaults to True.
            conventional_unit_cell (bool): Whether to get the standard
                conventional unit cell for the final or list of initial structures.

        Returns:
            Structure object or list of Structure objects.
        """

        structure_data = self.materials.get_structure_by_material_id(  # type: ignore
            material_id=material_id, final=final
        )

        if conventional_unit_cell and structure_data:
            if final:
                structure_data = SpacegroupAnalyzer(
                    structure_data
                ).get_conventional_standard_structure()
            else:
                structure_data = [
                    SpacegroupAnalyzer(structure).get_conventional_standard_structure()
                    for structure in structure_data
                ]

        return structure_data

    def get_database_version(self):
        """
        The Materials Project database is periodically updated and has a
        database version associated with it. When the database is updated,
        consolidated data (information about "a material") may and does
        change, while calculation data about a specific calculation task
        remains unchanged and available for querying via its task_id.

        The database version is set as a date in the format YYYY_MM_DD,
        where "_DD" may be optional. An additional numerical suffix
        might be added if multiple releases happen on the same day.

        Returns: database version as a string
        """
        return BaseRester(endpoint=self.endpoint + "/heartbeat")._query_resource()[
            "db_version"
        ]

    def get_materials_id_from_task_id(self, task_id):
        """
        Returns the current material_id from a given task_id. The
        materials_id should rarely change, and is usually chosen from
        among the smallest numerical id from the group of task_ids for
        that material. However, in some circumstances it might change,
        and this method is useful for finding the new material_id.

        Args:
            task_id (str): A task id.

        Returns:
            materials_id (MPID)
        """
        docs = self.materials.search(task_ids=[task_id], fields=["material_id"])
        if len(docs) == 1:  # pragma: no cover
            return str(docs[0].material_id)
        elif len(docs) > 1:  # pragma: no cover
            raise ValueError(
                f"Multiple documents return for {task_id}, this should not happen, please report it!"
            )
        else:  # pragma: no cover
            warnings.warn(
                f"No material found containing task {task_id}. Please report it if you suspect a task has gone missing."
            )
            return None

    def get_materials_id_references(self, material_id):
        """
        Returns all references for a materials id.

        Args:
            material_id (str): A material id.

        Returns:
            BibTeX (str)
        """
        raise NotImplementedError

    def get_materials_ids(self, chemsys_formula):
        """
        Get all materials ids for a formula or chemsys.

        Args:
            chemsys_formula (str): A chemical system (e.g., Li-Fe-O),
                or formula (e.g., Fe2O3).

        Returns:
            ([MPID]) List of all materials ids.
        """
        return sorted(
            doc.material_id
            for doc in self.materials.search_material_docs(
                chemsys_formula=chemsys_formula,
                all_fields=False,
                fields=["material_id"],
            )
        )

    def get_structures(self, chemsys_formula, final=True):
        """
        Get a list of Structures corresponding to a chemical system, formula,
        or materials_id.

        Args:
            chemsys_formula_id (str): A chemical system (e.g., Li-Fe-O),
                or formula (e.g., Fe2O3).
            final (bool): Whether to get the final structure, or the list of initial
                (pre-relaxation) structures. Defaults to True.

        Returns:
            List of Structure objects.
        """

        if final:
            return [
                doc.structure
                for doc in self.materials.search_material_docs(
                    chemsys_formula=chemsys_formula,
                    all_fields=False,
                    fields=["structure"],
                )
            ]
        else:
            structures = []

            for doc in self.materials.search_material_docs(
                chemsys_formula=chemsys_formula,
                all_fields=False,
                fields=["initial_structures"],
            ):
                structures.extend(doc.initial_structures)

            return structures

    def get_entries(
        self, chemsys_formula, sort_by_e_above_hull=False,
    ):
        """
        Get a list of ComputedEntries or ComputedStructureEntries corresponding
        to a chemical system or formula.

        Args:
            chemsys_formula (str): A chemical system
                (e.g., Li-Fe-O), or formula (e.g., Fe2O3).
            sort_by_e_above_hull (bool): Whether to sort the list of entries by
                e_above_hull in ascending order.

        Returns:
            List of ComputedEntry or ComputedStructureEntry objects.
        """

        entries = []

        if sort_by_e_above_hull:

            for doc in self.thermo.search_thermo_docs(
                chemsys_formula=chemsys_formula,
                all_fields=False,
                fields=["entries"],
                sort_field="energy_above_hull",
                ascending=True,
            ):
                entries.extend(list(doc.entries.values()))

            return entries

        else:
            for doc in self.thermo.search_thermo_docs(
                chemsys_formula=chemsys_formula, all_fields=False, fields=["entries"],
            ):
                entries.extend(list(doc.entries.values()))

            return entries

    def get_entry_by_material_id(self, material_id):
        """
        Get all ComputedEntry objects corresponding to a material_id.

        Args:
            material_id (str): Materials Project material_id (a string,
                e.g., mp-1234).
        Returns:
            List of ComputedEntry or ComputedStructureEntry object.
        """
        return list(
            self.thermo.get_document_by_id(
                document_id=material_id, fields=["entries"]
            ).entries.values()
        )

    def get_phonon_dos_by_material_id(self, material_id):
        """
        Get phonon density of states data corresponding to a material_id.

        Args:
            material_id (str): Materials Project material_id.

        Returns:
             CompletePhononDos: A phonon DOS object.

        """
        raise NotImplementedError

    def get_phonon_bandstructure_by_material_id(self, material_id):
        """
        Get phonon dispersion data corresponding to a material_id.

        Args:
            material_id (str): Materials Project material_id.

        Returns:
            PhononBandStructureSymmLine:  phonon band structure.
        """
        doc = self.phonon.get_document_by_id(material_id)

        return doc.ph_bs

    def query(
        self,
        material_ids: Optional[List[MPID]] = None,
        chemsys_formula: Optional[str] = None,
        nsites: Optional[Tuple[int, int]] = None,
        volume: Optional[Tuple[float, float]] = None,
        density: Optional[Tuple[float, float]] = None,
        crystal_system: Optional[CrystalSystem] = None,
        spacegroup_number: Optional[int] = None,
        spacegroup_symbol: Optional[str] = None,
        deprecated: Optional[bool] = None,
        total_energy: Optional[Tuple[float, float]] = None,
        formation_energy: Optional[Tuple[float, float]] = None,
        energy_above_hull: Optional[Tuple[float, float]] = None,
        equillibrium_reaction_energy: Optional[Tuple[float, float]] = None,
        uncorrected_energy: Optional[Tuple[float, float]] = None,
        is_stable: Optional[bool] = None,
        band_gap: Optional[Tuple[float, float]] = None,
        efermi: Optional[Tuple[float, float]] = None,
        is_gap_direct: Optional[bool] = None,
        is_metal: Optional[bool] = None,
        magnetic_ordering: Optional[Ordering] = None,
        total_magnetization: Optional[Tuple[float, float]] = None,
        total_magnetization_normalized_vol: Optional[Tuple[float, float]] = None,
        total_magnetization_normalized_formula_units: Optional[
            Tuple[float, float]
        ] = None,
        k_voigt: Optional[Tuple[float, float]] = None,
        k_reuss: Optional[Tuple[float, float]] = None,
        k_vrh: Optional[Tuple[float, float]] = None,
        g_voigt: Optional[Tuple[float, float]] = None,
        g_reuss: Optional[Tuple[float, float]] = None,
        g_vrh: Optional[Tuple[float, float]] = None,
        elastic_anisotropy: Optional[Tuple[float, float]] = None,
        poisson_ratio: Optional[Tuple[float, float]] = None,
        e_total: Optional[Tuple[float, float]] = None,
        e_ionic: Optional[Tuple[float, float]] = None,
        e_static: Optional[Tuple[float, float]] = None,
        n: Optional[Tuple[float, float]] = None,
        piezoelectric_modulus: Optional[Tuple[float, float]] = None,
        weighted_surface_energy: Optional[Tuple[float, float]] = None,
        weighted_work_function: Optional[Tuple[float, float]] = None,
        surface_anisotropy: Optional[Tuple[float, float]] = None,
        shape_factor: Optional[Tuple[float, float]] = None,
        has_props: Optional[List[str]] = None,
        theoretical: Optional[bool] = None,
        sort_field: Optional[str] = None,
        ascending: Optional[bool] = None,
        num_chunks: Optional[int] = None,
        chunk_size: int = 1000,
        all_fields: bool = True,
        fields: Optional[List[str]] = None,
    ):
        """
        Query core data using a variety of search criteria.

        Arguments:
            material_ids (List[MPID]): List of Materials Project IDs to return data for.
            chemsys_formula (str): A chemical system (e.g., Li-Fe-O),
                or formula including anonomyzed formula
                or wild cards (e.g., Fe2O3, ABO3, Si*).
            crystal_system (CrystalSystem): Crystal system of material.
            spacegroup_number (int): Space group number of material.
            spacegroup_symbol (str): Space group symbol of the material in international short symbol notation.
            nsites (Tuple[int,int]): Minimum and maximum number of sites to consider.
            volume (Tuple[float,float]): Minimum and maximum volume to consider.
            density (Tuple[float,float]): Minimum and maximum density to consider.
            deprecated (bool): Whether the material is tagged as deprecated.
            total_energy (Tuple[int,int]): Minimum and maximum corrected total energy in eV/atom to consider.
            formation_energy (Tuple[int,int]): Minimum and maximum formation energy in eV/atom to consider.
            energy_above_hull (Tuple[int,int]): Minimum and maximum energy above the hull in eV/atom to consider.
            uncorrected_energy (Tuple[int,int]): Minimum and maximum uncorrected total energy in eV/atom to consider.
            band_gap (Tuple[float,float]): Minimum and maximum band gap in eV to consider.
            efermi (Tuple[float,float]): Minimum and maximum fermi energy in eV to consider.
            is_gap_direct (bool): Whether the material has a direct band gap.
            is_metal (bool): Whether the material is considered a metal.
            magnetic_ordering (Ordering): Magnetic ordering of the material.
            total_magnetization (Tuple[float,float]): Minimum and maximum total magnetization values to consider.
            total_magnetization_normalized_vol (Tuple[float,float]): Minimum and maximum total magnetization values
                normalized by volume to consider.
            total_magnetization_normalized_formula_units (Tuple[float,float]): Minimum and maximum total magnetization
                values normalized by formula units to consider.
            k_voigt (Tuple[float,float]): Minimum and maximum value in GPa to consider for
                the Voigt average of the bulk modulus.
            k_reuss (Tuple[float,float]): Minimum and maximum value in GPa to consider for
                the Reuss average of the bulk modulus.
            k_vrh (Tuple[float,float]): Minimum and maximum value in GPa to consider for
                the Voigt-Reuss-Hill average of the bulk modulus.
            g_voigt (Tuple[float,float]): Minimum and maximum value in GPa to consider for
                the Voigt average of the shear modulus.
            g_reuss (Tuple[float,float]): Minimum and maximum value in GPa to consider for
                the Reuss average of the shear modulus.
            g_vrh (Tuple[float,float]): Minimum and maximum value in GPa to consider for
                the Voigt-Reuss-Hill average of the shear modulus.
            elastic_anisotropy (Tuple[float,float]): Minimum and maximum value to consider for
                the elastic anisotropy.
            poisson_ratio (Tuple[float,float]): Minimum and maximum value to consider for
                Poisson's ratio.
            e_total (Tuple[float,float]): Minimum and maximum total dielectric constant to consider.
            e_ionic (Tuple[float,float]): Minimum and maximum ionic dielectric constant to consider.
            e_static (Tuple[float,float]): Minimum and maximum electronic dielectric constant to consider.
            n (Tuple[float,float]): Minimum and maximum refractive index to consider.
            piezoelectric_modulus (Tuple[float,float]): Minimum and maximum piezoelectric modulus to consider.
            weighted_surface_energy (Tuple[float,float]): Minimum and maximum weighted surface energy in J/m² to
                consider.
            weighted_work_function (Tuple[float,float]): Minimum and maximum weighted work function in eV to consider.
            surface_energy_anisotropy (Tuple[float,float]): Minimum and maximum surface energy anisotropy values to
                consider.
            shape_factor (Tuple[float,float]): Minimum and maximum shape factor values to consider.
            has_props: (List[str]): The calculated properties available for the material.
            theoretical: (bool): Whether the material is theoretical.
            sort_field (str): Field used to sort results.
            ascending (bool): Whether sorting should be in ascending order.
            num_chunks (int): Maximum number of chunks of data to yield. None will yield all possible.
            chunk_size (int): Number of data entries per chunk.
            all_fields (bool): Whether to return all fields in the document. Defaults to True.
            fields (List[str]): List of fields in SearchDoc to return data for.
                Default is material_id if all_fields is False.

        Returns:
            ([SearchDoc]) List of SearchDoc documents
        """
        return self.search.search_docs(  # type: ignore
            material_ids=material_ids,
            chemsys_formula=chemsys_formula,
            nsites=nsites,
            volume=volume,
            density=density,
            crystal_system=crystal_system,
            spacegroup_number=spacegroup_number,
            spacegroup_symbol=spacegroup_symbol,
            deprecated=deprecated,
            total_energy=total_energy,
            formation_energy=formation_energy,
            energy_above_hull=energy_above_hull,
            equillibrium_reaction_energy=equillibrium_reaction_energy,
            uncorrected_energy=uncorrected_energy,
            is_stable=is_stable,
            band_gap=band_gap,
            efermi=efermi,
            is_gap_direct=is_gap_direct,
            is_metal=is_metal,
            magnetic_ordering=magnetic_ordering,
            total_magnetization=total_magnetization,
            total_magnetization_normalized_vol=total_magnetization_normalized_vol,
            total_magnetization_normalized_formula_units=total_magnetization_normalized_formula_units,
            k_voigt=k_voigt,
            k_reuss=k_reuss,
            k_vrh=k_vrh,
            g_voigt=g_voigt,
            g_reuss=g_reuss,
            g_vrh=g_vrh,
            elastic_anisotropy=elastic_anisotropy,
            poisson_ratio=poisson_ratio,
            e_total=e_total,
            e_ionic=e_ionic,
            e_static=e_static,
            n=n,
            piezoelectric_modulus=piezoelectric_modulus,
            weighted_surface_energy=weighted_surface_energy,
            weighted_work_function=weighted_work_function,
            surface_anisotropy=surface_anisotropy,
            shape_factor=shape_factor,
            has_props=has_props,
            theoretical=theoretical,
            sort_field=sort_field,
            ascending=ascending,
            num_chunks=num_chunks,
            chunk_size=chunk_size,
            all_fields=all_fields,
            fields=fields,
        )

    def submit_structures(self, structures, public_name, public_email):
        """
        Submits a list of structures to the Materials Project.

        Note that public_name and public_email will be used to credit the
        submitter on the Materials Project website.

        Args:
            structures: A list of Structure objects

        Returns:
            ?
        """
        # TODO: call new MPComplete endpoint
        raise NotImplementedError

    def get_stability(self, entries):
        """
        Returns the stability of all entries.
        """
        # TODO: discuss
        raise NotImplementedError

    def get_cohesive_energy(self, material_id, per_atom=False):
        """
        Gets the cohesive for a material (eV per formula unit). Cohesive energy
            is defined as the difference between the bulk energy and the sum of
            total DFT energy of isolated atoms for atom elements in the bulk.
        Args:
            material_id (str): Materials Project material_id, e.g. 'mp-123'.
            per_atom (bool): Whether or not to return cohesive energy per atom
        Returns:
            Cohesive energy (eV).
        """
        raise NotImplementedError

    def get_reaction(self, reactants, products):
        """
        Gets a reaction from the Materials Project.

        Args:
            reactants ([str]): List of formulas
            products ([str]): List of formulas

        Returns:
            rxn
        """
        raise NotImplementedError

    def get_substrates(self, material_id, orient=None):
        """
        Get a substrate list for a material id. The list is in order of
        increasing elastic energy if a elastic tensor is available for
        the material_id.

        Args:
            material_id (str): Materials Project material_id, e.g. 'mp-123'.
            orient (list) : substrate orientation to look for
        Returns:
            list of dicts with substrate matches
        """

        return [
            doc.dict()
            for doc in self.substrates.search_substrates_docs(
                film_id=material_id,
                substrate_orientation=orient,
                sort_field="energy",
                ascending=True,
            )
        ]

    def get_surface_data(self, material_id, miller_index=None):
        """
        Gets surface data for a material. Useful for Wulff shapes.

        Reference for surface data:

        Tran, R., Xu, Z., Radhakrishnan, B., Winston, D., Sun, W., Persson, K.
        A., & Ong, S. P. (2016). Data Descripter: Surface energies of elemental
        crystals. Scientific Data, 3(160080), 1–13.
        http://dx.doi.org/10.1038/sdata.2016.80

        Args:
            material_id (str): Materials Project material_id, e.g. 'mp-123'.
            miller_index (list of integer): The miller index of the surface.
            e.g., [3, 2, 1]. If miller_index is provided, only one dictionary
            of this specific plane will be returned.

        Returns:
            Surface data for material. Energies are given in SI units (J/m^2).
        """
        doc = self.surface_properties.get_document_by_id(material_id)
        structure = self.get_structure_by_material_id(material_id)

        if miller_index:
            eq_indices = get_symmetrically_equivalent_miller_indices(
                structure, miller_index
            )

            for surface in doc.surfaces:
                if tuple(surface.miller_index) in eq_indices:
                    return surface.dict()
        else:
            return doc.dict()

    def get_wulff_shape(self, material_id):
        """
        Constructs a Wulff shape for a material.

        Args:
            material_id (str): Materials Project material_id, e.g. 'mp-123'.
        Returns:
            pymatgen.analysis.wulff.WulffShape
        """
        from pymatgen.analysis.wulff import WulffShape
        from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

        structure = self.get_structure_by_material_id(material_id)
        surfaces = self.get_surface_data(material_id)["surfaces"]
        lattice = (
            SpacegroupAnalyzer(structure).get_conventional_standard_structure().lattice
        )
        miller_energy_map = {}
        for surf in surfaces:
            miller = tuple(surf["miller_index"])
            # Prefer reconstructed surfaces, which have lower surface energies.
            if (miller not in miller_energy_map) or surf["is_reconstructed"]:
                miller_energy_map[miller] = surf["surface_energy"]
        millers, energies = zip(*miller_energy_map.items())
        return WulffShape(lattice, millers, energies)

    def get_gb_data(
        self,
        material_id=None,
        pretty_formula=None,
        chemsys=None,
        sigma=None,
        gb_plane=None,
        rotation_axis=None,
    ):
        """
        Gets grain boundary data for a material.

        Args:
            material_id (str): Materials Project material_id, e.g., 'mp-129'.
            pretty_formula (str): The formula of metals. e.g., 'Fe'
            chemsys (str): Dash delimited elements in material.
            sigma(int): The sigma value of a certain type of grain boundary
            gb_plane(list of integer): The Miller index of grain
            boundary plane. e.g., [1, 1, 1]
            rotation_axis(list of integer): The Miller index of rotation
            axis. e.g., [1, 0, 0], [1, 1, 0], and [1, 1, 1]
            Sigma value is determined by the combination of rotation axis and
            rotation angle. The five degrees of freedom (DOF) of one grain boundary
            include: rotation axis (2 DOFs), rotation angle (1 DOF), and grain
            boundary plane (2 DOFs).
            include_work_of_separation (bool): whether to include the work of separation
            (in unit of (J/m^2)). If you want to query the work of separation, please
            specify the material_id.


        Returns:
            A list of grain boundaries that satisfy the query conditions (sigma, gb_plane).
            Energies are given in SI units (J/m^2).
        """
        return [
            doc.dict()
            for doc in self.grain_boundary.search_grain_boundary_docs(
                material_ids=[material_id] if material_id else None,
                pretty_formula=pretty_formula,
                chemsys=chemsys,
                sigma=sigma,
                gb_plane=gb_plane,
                rotation_axis=rotation_axis,
            )
        ]

    def get_interface_reactions(
        self,
        reactant1,
        reactant2,
        open_el=None,
        relative_mu=None,
        use_hull_energy=False,
    ):
        """
        Gets critical reactions between two reactants.

        Get critical reactions ("kinks" in the mixing ratio where
        reaction products change) between two reactants. See the
        `pymatgen.analysis.interface_reactions` module for more info.

        Args:
            reactant1 (str): Chemical formula for reactant
            reactant2 (str): Chemical formula for reactant
            open_el (str): Element in reservoir available to system
            relative_mu (float): Relative chemical potential of element in
                reservoir with respect to pure substance. Must be non-positive.
            use_hull_energy (bool): Whether to use the convex hull energy for a
            given composition for the reaction energy calculation. If false,
            the energy of the ground state structure will be preferred; if a
            ground state can not be found for a composition, the convex hull
            energy will be used with a warning message.

        Returns:
            list: list of dicts of form {ratio,energy,rxn} where `ratio` is the
                reactant mixing ratio, `energy` is the reaction energy
                in eV/atom, and `rxn` is a
                `pymatgen.analysis.reaction_calculator.Reaction`.

        """
        raise NotImplementedError



from typing import Dict, List
from maggma.api.resource.core import Resource
from pymatgen.core import __version__ as pmg_version  # type: ignore
from mp_api.core.settings import MAPISettings
from fastapi.openapi.utils import get_openapi
from maggma.api.API import API


class MAPI(API):
    """
    Core Materials API that orchestrates resources together
    """

    def __init__(
        self,
        resources: Dict[str, List[Resource]],
        title="Materials Project API",
        version="v3.0-dev",
        debug=False,
        heartbeat_meta={"pymatgen": pmg_version, "db_version": MAPISettings().db_version},
    ):
        super().__init__(
            resources=resources, title=title, version=version, debug=debug, heartbeat_meta=heartbeat_meta,
        )

    @property
    def app(self):
        """
        App server for the cluster manager
        """
        app = super().app

        def custom_openapi():
            openapi_schema = get_openapi(title=self.title, version=self.version, routes=app.routes)

            openapi_schema["components"]["securitySchemes"] = {
                "ApiKeyAuth": {
                    "descriptions": "MP API key to authorize requests",
                    "name": "X-API-KEY",
                    "in": "header",
                    "type": "apiKey",
                }
            }

            openapi_schema["security"] = [{"ApiKeyAuth": []}]

            openapi_schema["info"]["x-logo"] = {
                "url": "https://raw.githubusercontent.com/materialsproject/api/master/src/mp_api/core/assets/mp_logo_small.png"  # noqa: E501
            }

            app.openapi_schema = openapi_schema
            return app.openapi_schema

        app.openapi = custom_openapi

        return app



from typing import Dict, List, Optional, Union

from emmet.core.electronic_structure import BandstructureData, DosData
from emmet.core.mpid import MPID
from emmet.core.symmetry import SymmetryData
from emmet.core.thermo import DecompositionProduct
from emmet.core.xas import Edge, Type
from mp_api.routes.grain_boundary.models import GBTypeEnum
from pydantic import BaseModel, Field
from pymatgen.analysis.magnetism.analyzer import Ordering
from pymatgen.core.composition import Composition
from pymatgen.core.periodic_table import Element
from pymatgen.core.structure import Structure


class SearchStats(BaseModel):
    """
    Statistics about a specified SearchDoc field.
    """

    field: str = Field(
        None,
        title="Field",
        description="Field name corresponding to a field in SearchDoc",
    )
    num_samples: Optional[int] = Field(
        None,
        title="Sample",
        description="The number of documents sampled to generate statistics. "
        "If unspecified, statistics will be from entire database.",
    )
    min: float = Field(
        None,
        title="Minimum",
        description="The minimum value "
        "of the specified field used to "
        "generate statistics.",
    )
    max: float = Field(
        None,
        title="Maximum",
        description="The maximum value "
        "of the specified field used to "
        "generate statistics.",
    )
    median: float = Field(
        None, title="Median", description="The median of the field values."
    )
    mean: float = Field(None, title="Mean", description="The mean of the field values.")
    distribution: List[float] = Field(
        None,
        title="Distribution",
        description="List of floats specifying a kernel density "
        "estimator of the distribution, equally spaced "
        "between specified minimum and maximum values.",
    )
    warnings: List[str] = Field(
        None,
        title="Warnings",
        description="Any warnings generated while generating statistics.",
    )


class XASSearchData(BaseModel):
    """
    Fields in XAS sub docs in search
    """

    edge: Edge = Field(
        None, title="Absorption Edge", description="The interaction edge for XAS"
    )
    absorbing_element: Element = Field(None, title="Absorbing Element")

    spectrum_type: Type = Field(None, title="Type of XAS Spectrum")


class GBSearchData(BaseModel):
    """
    Fields in grain boundary sub docs in search
    """

    sigma: int = Field(
        None, description="Sigma value of the boundary",
    )

    type: GBTypeEnum = Field(
        None, description="Grain boundary type",
    )

    gb_energy: float = Field(
        None, description="Grain boundary energy in J/m^2",
    )

    rotation_angle: float = Field(
        None, description="Rotation angle in degrees",
    )


class SearchDoc(BaseModel):
    """
    Summary information about materials and their properties, useful for materials
    screening studies and searching.
    """

    # Materials

    material_id: Union[MPID, int] = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******",
    )

    nsites: int = Field(None, description="Total number of sites in the structure")
    elements: List[Element] = Field(
        None, description="List of elements in the material"
    )
    nelements: int = Field(None, title="Number of Elements")
    composition: Composition = Field(
        None, description="Full composition for the material"
    )
    composition_reduced: Composition = Field(
        None,
        title="Reduced Composition",
        description="Simplified representation of the composition",
    )
    formula_pretty: str = Field(
        None,
        title="Pretty Formula",
        description="Cleaned representation of the formula",
    )
    formula_anonymous: str = Field(
        None,
        title="Anonymous Formula",
        description="Anonymized representation of the formula",
    )
    chemsys: str = Field(
        None,
        title="Chemical System",
        description="dash-delimited string of elements in the material",
    )
    volume: float = Field(
        None,
        title="Volume",
        description="Total volume for this structure in Angstroms^3",
    )

    density: float = Field(
        None, title="Density", description="Density in grams per cm^3"
    )

    symmetry: SymmetryData = Field(None, description="Symmetry data for this material")

    deprecated: bool = Field(
        None, description="Whether the material is tagged as deprecated"
    )

    structure: Structure = Field(
        None, description="The lowest energy structure for this material"
    )

    # Thermo

    uncorrected_energy_per_atom: float = Field(
        None, description="The total DFT energy of this material per atom in eV/atom"
    )

    energy_per_atom: float = Field(
        None,
        description="The total corrected DFT energy of this material per atom in eV/atom",
    )

    formation_energy_per_atom: float = Field(
        None, description="The formation energy per atom in eV/atom"
    )

    energy_above_hull: float = Field(
        None, description="The energy above the hull in eV/Atom"
    )

    is_stable: bool = Field(
        False,
        description="Flag for whether this material is on the hull and therefore stable",
    )

    equillibrium_reaction_energy_per_atom: float = Field(
        None,
        description="The reaction energy of a stable entry from the neighboring equilibrium stable materials in eV."
        " Also known as the inverse distance to hull.",
    )

    decomposes_to: List[DecompositionProduct] = Field(
        None,
        description="List of decomposition data for this material. Only valid for metastable or unstable material.",
    )

    # XAS

    xas: List[XASSearchData] = Field(
        None, description="List of xas documents.",
    )

    # GB

    grain_boundaries: List[GBSearchData] = Field(
        None, description="List of grain boundary documents.",
    )

    # Electronic Structure

    band_gap: float = Field(
        None, description="Band gap energy in eV.",
    )

    cbm: Union[float, Dict] = Field(
        None, description="Conduction band minimum data.",
    )

    vbm: Union[float, Dict] = Field(
        None, description="Valence band maximum data.",
    )

    efermi: float = Field(
        None, description="Fermi energy eV.",
    )

    is_gap_direct: bool = Field(
        None, description="Whether the band gap is direct.",
    )

    is_metal: bool = Field(
        None, description="Whether the material is a metal.",
    )

    magnetic_ordering: Union[str, Ordering] = Field(
        None, description="Magnetic ordering of the calculation.",
    )

    es_source_calc_id: Union[MPID, int] = Field(
        None, description="The source calculation ID for the electronic structure data."
    )

    bandstructure: BandstructureData = Field(
        None, description="Band structure data for the material."
    )

    dos: DosData = Field(None, description="Density of states data for the material.")

    # DOS

    dos_energy_up: float = Field(
        None, description="Spin-up DOS band gap.",
    )

    dos_energy_down: float = Field(
        None, description="Spin-down DOS band gap.",
    )

    # Magnetism

    ordering: str = Field(
        None, description="Type of magnetic ordering.",
    )

    total_magnetization: float = Field(
        None, description="Total magnetization in μB.",
    )

    total_magnetization_normalized_vol: float = Field(
        None, description="Total magnetization normalized by volume in μB/Å³.",
    )

    total_magnetization_normalized_formula_units: float = Field(
        None, description="Total magnetization normalized by formula unit in μB/f.u. .",
    )

    # Elasticity

    k_voigt: float = Field(
        None, description="Voigt average of the bulk modulus.",
    )

    k_reuss: float = Field(
        None, description="Reuss average of the bulk modulus in GPa.",
    )

    k_vrh: float = Field(
        None, description="Voigt-Reuss-Hill average of the bulk modulus in GPa.",
    )

    g_voigt: float = Field(
        None, description="Voigt average of the shear modulus in GPa.",
    )

    g_reuss: float = Field(
        None, description="Reuss average of the shear modulus in GPa.",
    )

    g_vrh: float = Field(
        None, description="Voigt-Reuss-Hill average of the shear modulus in GPa.",
    )

    universal_anisotropy: float = Field(
        None, description="Elastic anisotropy.",
    )

    homogeneous_poisson: float = Field(
        None, description="Poisson's ratio.",
    )

    # Dielectric and Piezo

    e_total: float = Field(
        None, description="Total dielectric constant",
    )

    e_ionic: float = Field(
        None, description="Ionic contributio to dielectric constant",
    )

    e_static: float = Field(
        None, description="Electronic contribution to dielectric constant",
    )

    n: float = Field(
        None, description="Refractive index",
    )

    e_ij_max: float = Field(
        None, description="Piezoelectric modulus",
    )

    # Surface Properties

    weighted_surface_energy_EV_PER_ANG2: float = Field(
        None, description="Weighted surface energy in eV/Å²",
    )

    weighted_surface_energy: float = Field(
        None, description="Weighted surface energy in J/m²",
    )

    weighted_work_function: float = Field(
        None, description="Weighted work function in eV.",
    )

    surface_anisotropy: float = Field(
        None, description="Surface energy anisotropy.",
    )

    shape_factor: float = Field(
        None, description="Shape factor.",
    )

    # Has Props

    has_props: List[str] = Field(
        None, description="List of properties that are available for a given material.",
    )

    # Theoretical

    theoretical: bool = Field(
        None, description="Whether the material is theoretical.",
    )



from datetime import datetime
from monty.json import MontyDecoder

from emmet.core.polar import Dielectric

from pydantic import BaseModel, Field, validator


class DielectricDoc(BaseModel):
    """
    Dielectric tensor and associated information.
    """

    dielectric: Dielectric = Field(
        None, description="Dielectric data",
    )

    task_id: str = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******",
    )

    last_updated: datetime = Field(
        None,
        description="Timestamp for the most recent calculation for this Material document",
    )

    # Make sure that the datetime field is properly formatted
    @validator("last_updated", pre=True)
    def last_updated_dict_ok(cls, v):
        return MontyDecoder().process_decoded(v)



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field


class Value(BaseModel):
    min_value: float = Field(
        None,
        description="Minimal value."
    )
    max_value: float = Field(
        None,
        description="Maximal value."
    )
    values: List[float] = Field(
        [],
        description="Enumerated values in the literature."
    )
    units: str = Field(
        ...,
        description="Unit of this value."
    )


class Conditions(BaseModel):
    heating_temperature: Optional[List[Value]] = Field(
        None,
        description="Heating temperatures."
    )
    heating_time: Optional[List[Value]] = Field(
        None,
        description="Heating times."
    )
    heating_atmosphere: Optional[List[str]] = Field(
        None,
        description="List of heating atmospheres."
    )
    mixing_device: Optional[str] = Field(
        None,
        description="Mixing device, if this operation is MIXING."
    )
    mixing_media: Optional[str] = Field(
        None,
        description="Mixing media, if this operation is MIXING."
    )


class OperationTypeEnum(str, Enum):
    starting = "StartingSynthesis"
    mixing = "MixingOperation"
    shaping = "ShapingOperation"
    drying = "DryingOperation"
    heating = "HeatingOperation"
    quenching = "QuenchingOperation"


class Operation(BaseModel):
    type: OperationTypeEnum = Field(
        ...,
        description="Type of the operation as classified by the pipeline."
    )
    token: str = Field(
        ...,
        description="Token (word) of the operation as written in paper."
    )
    conditions: Conditions = Field(
        ...,
        description="The conditions linked to this operation."
    )



import pytest
from mp_api.routes.synthesis.client import SynthesisRester

import inspect


@pytest.fixture
def rester():
    rester = SynthesisRester()
    yield rester
    rester.session.close()


def test_client(rester):
    # Get specific search method
    search_method = None
    for entry in inspect.getmembers(rester, predicate=inspect.ismethod):
        if "search" in entry[0] and entry[0] != "search":
            search_method = entry[1]

    if search_method is not None:

        q = {"keywords": ["silicon"]}

        doc = search_method(**q)[0]

        assert doc.doi is not None
        assert doc.paragraph_string is not None
        assert doc.synthesis_type is not None




from mp_api.routes._consumer.models import UserSettingsDoc
from mp_api.core.client import BaseRester


class UserSettingsRester(BaseRester):  # pragma: no cover

    suffix = "user_settings"
    document_model = UserSettingsDoc  # type: ignore

    def set_user_settings(self, consumer_id, settings):  # pragma: no cover
        """
        Set user settings.
        Args:
            consumer_id: Consumer ID for the user
            settings: Dictionary with user settings
        Returns:
            Dictionary with consumer_id and write status.
        Raises:
            MPRestError
        """
        return self._post_resource(
            body=settings, params={"consumer_id": consumer_id}
        ).get("data")

    def get_user_settings(self, consumer_id, settings):  # pragma: no cover
        """
        Get user settings.
        Args:
            consumer_id: Consumer ID for the user
        Returns:
            Dictionary with consumer_id and settings.
        Raises:
            MPRestError
        """
        return self.query(query={"consumer_id": consumer_id}, monty_decode=False)



"""
This script converts synthesis recipes data fetched directly
from the public repo of synthesis recipes
(https://github.com/CederGroupHub/text-mined-synthesis_public)
into MP compatible formats.
"""
import json
import sys

from pymatgen.core import Composition
from pymatgen.core.composition import CompositionError


def string2comp(x):
    """Convert string material formulas into pymatgen Compositions."""
    # TODO: if a material contains multiple parts, this function
    #  only takes the first part. This is not the optimal solution,
    #  and should be resolved in the future.
    formula = x.split('·')

    # return reduce(add, [Composition(x) for x in formula])
    return Composition(formula[0])


def convert_recipe(recipe):
    """Convert an entire synthesis recipe."""
    targets_string = recipe['targets_string']
    try:
        target_comps = [string2comp(x) for x in targets_string]
    except (CompositionError, ValueError):
        print('Cannot process materials: ', targets_string)
        raise

    recipe['targets_formula'] = [json.loads(x.to_json()) for x in target_comps]
    recipe['targets_formula_s'] = [x.reduced_formula for x in target_comps]
    del recipe['targets_string']

    recipe['precursors_formula'] = []
    recipe['precursors_formula_s'] = []
    for precursor in recipe['precursors']:
        try:
            comp = string2comp(precursor['material_formula'])
        except (CompositionError, ValueError):
            print('Cannot process precursor material: ', precursor['material_formula'])
            continue
        recipe['precursors_formula'].append(json.loads(comp.to_json()))
        recipe['precursors_formula_s'].append(comp.reduced_formula)

    return recipe


def convert_json_public_repo(src_json, dst_json):
    """
    Convert the public synthesis recipes dataset (in a json file)
    into a format as json file which can be imported into the MP database.
    """
    with open(src_json) as f:
        data = json.load(f)
        recipes = data['reactions']

        print('Loaded %s recipes, version %s' % (len(recipes), data['release_date']))

    converted = []
    for recipe in recipes:
        try:
            convert_recipe(recipe)
            converted.append(recipe)
        except (CompositionError, ValueError, IndexError):
            pass

    print('Converted %d recipes' % (len(converted),))
    with open(dst_json, 'w') as f:
        json.dump(converted, f)


if __name__ == '__main__':
    convert_json_public_repo(sys.argv[1], sys.argv[2])



from pydantic import BaseSettings, Field
from mp_api import __file__ as root_dir
import os


class MAPISettings(BaseSettings):
    """
    Special class to store settings for MAPI
    python module
    """

    app_path: str = Field(
        "~/mapi.json", description="Path for the default MAPI JSON definition"
    )

    debug: bool = Field(False, description="Turns on debug mode for MAPI")

    test_files: str = Field(
        os.path.join(os.path.dirname(os.path.abspath(root_dir)), "../../test_files"),
        description="Directory with test files",
    )

    db_version: str = Field("2021_04_26", description="Database version")

    requests_per_min: int = Field(
        60, description="Number of requests per minute to for rate limit."
    )

    class Config:
        env_prefix = "mapi_"



from maggma.api.resource import ReadOnlyResource
from mp_api.routes.similarity.models import SimilarityDoc
from maggma.api.query_operator import PaginationQuery, SparseFieldsQuery


def similarity_resource(similarity_store):
    resource = ReadOnlyResource(
        similarity_store,
        SimilarityDoc,
        query_operators=[
            PaginationQuery(),
            SparseFieldsQuery(SimilarityDoc, default_fields=["task_id"]),
        ],
        tags=["Similarity"],
        enable_default_search=False,
    )

    return resource



from typing import List, Optional, Tuple
from collections import defaultdict

from mp_api.core.client import BaseRester
from mp_api.routes.piezo.models import PiezoDoc

import warnings


class PiezoRester(BaseRester):

    suffix = "piezoelectric"
    document_model = PiezoDoc  # type: ignore
    primary_key = "task_id"

    def search_piezoelectric_docs(
        self,
        piezoelectric_modulus: Optional[Tuple[float, float]] = None,
        sort_field: Optional[str] = None,
        ascending: Optional[bool] = None,
        num_chunks: Optional[int] = None,
        chunk_size: int = 1000,
        all_fields: bool = True,
        fields: Optional[List[str]] = None,
    ):
        """
        Query equations of state docs using a variety of search criteria.

        Arguments:
            piezoelectric_modulus (Tuple[float,float]): Minimum and maximum of the
                piezoelectric modulus in C/m² to consider.
            sort_field (str): Field used to sort results.
            ascending (bool): Whether sorting should be in ascending order.
            num_chunks (int): Maximum number of chunks of data to yield. None will yield all possible.
            chunk_size (int): Number of data entries per chunk.
            all_fields (bool): Whether to return all fields in the document. Defaults to True.
            fields (List[str]): List of fields in PiezoDoc to return data for.
                Default is material_id and last_updated if all_fields is False.

        Returns:
            ([PiezoDoc]) List of piezoelectric documents
        """

        query_params = defaultdict(dict)  # type: dict

        if piezoelectric_modulus:
            query_params.update(
                {
                    "piezo_modulus_min": piezoelectric_modulus[0],
                    "piezo_modulus_max": piezoelectric_modulus[1],
                }
            )

        if sort_field:
            query_params.update({"sort_field": sort_field})

        if ascending is not None:
            query_params.update({"ascending": ascending})

        query_params = {
            entry: query_params[entry]
            for entry in query_params
            if query_params[entry] is not None
        }

        return super().search(
            num_chunks=num_chunks,
            chunk_size=chunk_size,
            all_fields=all_fields,
            fields=fields,
            **query_params
        )



import os
from json import load

from mp_api import MAPISettings
from mp_api.routes.tasks.utils import calcs_reversed_to_trajectory


def test_calcs_reversed_to_trajectory():
    with open(os.path.join(MAPISettings().test_files, "calcs_reversed_mp_1031016.json")) as file:
        calcs_reversed = load(file)
        trajectories = calcs_reversed_to_trajectory(calcs_reversed)

    assert len(trajectories) == 1
    assert trajectories[0]["lattice"] == [[9.054455, 0.0, 0.0], [0.0, 4.500098, 0.0], [0.0, 0.0, 4.500098]]



from typing import List, Dict, Optional

from pydantic import BaseModel, Field

__all__ = [
    'Component',
    'ExtractedMaterial'
]


class Component(BaseModel):
    formula: str = Field(
        ...,
        description="Formula of this component."
    )
    amount: str = Field(
        ...,
        description="Amount of this component."
    )
    elements: Dict[str, str] = Field(
        ...,
        description="Amount of each chemical elements in this component."
    )


class Values(BaseModel):
    values: List[float] = Field(
        None,
        description="List of values."
    )
    min_value: Optional[float] = Field(
        None,
        description="Minimal value."
    )
    max_value: Optional[float] = Field(
        None,
        description="Maximal value."
    )


class ExtractedMaterial(BaseModel):
    """
    Model for a material extracted from the literature
    """
    material_string: str = Field(
        ...,
        description="String of the material as written in paper."
    )
    material_formula: str = Field(
        ...,
        description="Normalized formula of the material."
    )
    material_name: str = Field(
        None,
        description="English name of the material."
    )

    phase: Optional[str] = Field(
        None,
        description="Phase description of material, such as anatase."
    )
    is_acronym: bool = Field(
        None,
        description="Whether the material is an acronym, such as LMO for LiMn2O4."
    )

    composition: List[Component] = Field(
        ...,
        description="List of components in this material."
    )
    amounts_vars: Dict[str, Values] = Field(
        {},
        description="Amount variables (formula subscripts)."
    )
    elements_vars: Dict[str, List[str]] = Field(
        {},
        description="Chemical element variables"
    )

    additives: List[str] = Field(
        [],
        description="List of additives, dopants, etc."
    )
    oxygen_deficiency: Optional[str] = Field(
        None,
        description="Symbol indicating whether the materials is oxygen deficient."
    )



from typing import List, Optional, Tuple
from collections import defaultdict

from mp_api.core.client import BaseRester
from mp_api.routes.dielectric.models import DielectricDoc


class DielectricRester(BaseRester):

    suffix = "dielectric"
    document_model = DielectricDoc  # type: ignore
    primary_key = "task_id"

    def search_dielectric_docs(
        self,
        e_total: Optional[Tuple[float, float]] = None,
        e_ionic: Optional[Tuple[float, float]] = None,
        e_static: Optional[Tuple[float, float]] = None,
        n: Optional[Tuple[float, float]] = None,
        sort_field: Optional[str] = None,
        ascending: Optional[bool] = None,
        num_chunks: Optional[int] = None,
        chunk_size: int = 1000,
        all_fields: bool = True,
        fields: Optional[List[str]] = None,
    ):
        """
        Query dielectric docs using a variety of search criteria.

        Arguments:
            e_total (Tuple[float,float]): Minimum and maximum total dielectric constant to consider.
            e_ionic (Tuple[float,float]): Minimum and maximum ionic dielectric constant to consider.
            e_static (Tuple[float,float]): Minimum and maximum electronic dielectric constant to consider.
            n (Tuple[float,float]): Minimum and maximum refractive index to consider.
            sort_field (str): Field used to sort results.
            ascending (bool): Whether sorting should be in ascending order.
            num_chunks (int): Maximum number of chunks of data to yield. None will yield all possible.
            chunk_size (int): Number of data entries per chunk.
            all_fields (bool): Whether to return all fields in the document. Defaults to True.
            fields (List[str]): List of fields in DielectricDoc to return data for.
                Default is material_id and last_updated if all_fields is False.

        Returns:
            ([DielectricDoc]) List of dielectric documents.
        """

        query_params = defaultdict(dict)  # type: dict

        if e_total:
            query_params.update({"e_total_min": e_total[0], "e_total_max": e_total[1]})

        if e_ionic:
            query_params.update({"e_ionic_min": e_ionic[0], "e_ionic_max": e_ionic[1]})

        if e_static:
            query_params.update({"e_static_min": e_static[0], "e_static_max": e_static[1]})

        if n:
            query_params.update({"n_min": n[0], "n_max": n[1]})

        if sort_field:
            query_params.update({"sort_field": sort_field})

        if ascending is not None:
            query_params.update({"ascending": ascending})

        query_params = {entry: query_params[entry] for entry in query_params if query_params[entry] is not None}

        return super().search(
            num_chunks=num_chunks, chunk_size=chunk_size, all_fields=all_fields, fields=fields, **query_params
        )



from emmet.core.symmetry import CrystalSystem
import pytest
import random
import os
import typing
from mp_api.matproj import MPRester
from mp_api.core.settings import MAPISettings
import requests

from pymatgen.io.cif import CifParser
from pymatgen.electronic_structure.bandstructure import (
    BandStructure,
    BandStructureSymmLine,
)
from pymatgen.electronic_structure.dos import CompleteDos
from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.core.periodic_table import Element
from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine
from pymatgen.io.vasp import Incar, Chgcar
from pymatgen.analysis.magnetism import Ordering
from pymatgen.analysis.wulff import WulffShape

api_is_up = (
    requests.get("https://api.materialsproject.org/heartbeat").status_code == 200
)


@pytest.fixture()
def mpr():
    rester = MPRester()
    yield rester
    rester.session.close()


@pytest.mark.skipif(
    ((os.environ.get("MP_API_KEY", None) is None) or (not api_is_up)),
    reason="API is down",
)
class TestMPRester:
    def test_get_structure_by_material_id(self, mpr):
        s1 = mpr.get_structure_by_material_id("mp-1")
        assert s1.formula == "Cs2"

        s1 = mpr.get_structure_by_material_id("mp-4163", conventional_unit_cell=True)
        assert s1.formula == "Ca12 Ti8 O28"

        s1 = mpr.get_structure_by_material_id("mp-149", final=False)
        assert [s.formula for s in s1] == ["Si2"]

        # # requesting via task-id instead of mp-id
        with pytest.warns(UserWarning):
            mpr.get_structure_by_material_id("mp-698856")

    def test_get_database_version(self, mpr):
        db_version = mpr.get_database_version()
        assert db_version == MAPISettings().db_version

    def test_get_materials_id_from_task_id(self, mpr):
        assert mpr.get_materials_id_from_task_id("mp-540081") == "mp-19017"

    # TODO: add method to MPRester
    # def test_get_materials_id_references(self, mpr):
    #     data = mpr.get_materials_id_references("mp-123")
    #     assert len(data) > 1000

    def test_get_materials_ids_doc(self, mpr):
        mpids = mpr.get_materials_ids("Al2O3")
        random.shuffle(mpids)
        doc = mpr.materials.get_document_by_id(mpids.pop(0))
        assert doc.formula_pretty == "Al2O3"

    def test_get_structures(self, mpr):
        structs = mpr.get_structures("Mn3O4")
        assert len(structs) > 0

        structs = mpr.get_structures("Mn3O4", final=False)
        assert len(structs) > 0

    def test_find_structure(self, mpr):
        path = os.path.join(MAPISettings().test_files, "Si_mp_149.cif")
        with open(path) as file:
            data = mpr.find_structure(path)
            assert len(data) > 0

            s = CifParser(file).get_structures()[0]
            data = mpr.find_structure(s)
            assert len(data) > 0

    def test_get_bandstructure_by_material_id(self, mpr):
        bs = mpr.get_bandstructure_by_material_id("mp-149")
        assert isinstance(bs, BandStructureSymmLine)
        bs_unif = mpr.get_bandstructure_by_material_id("mp-149", line_mode=False)
        assert isinstance(bs_unif, BandStructure)
        assert not isinstance(bs_unif, BandStructureSymmLine)

    def test_get_dos_by_id(self, mpr):
        dos = mpr.get_dos_by_material_id("mp-149")
        assert isinstance(dos, CompleteDos)

    def test_get_entry_by_material_id(self, mpr):
        e = mpr.get_entry_by_material_id("mp-19017")
        assert isinstance(e[0], ComputedEntry)
        assert e[0].composition.reduced_formula == "LiFePO4"

    def test_get_entries(self, mpr):
        syms = ["Li", "Fe", "O"]
        chemsys = "Li-Fe-O"
        entries = mpr.get_entries(chemsys)
        sorted_entries = mpr.get_entries(chemsys, sort_by_e_above_hull=True)

        elements = set([Element(sym) for sym in syms])
        for e in entries:
            assert isinstance(e, ComputedEntry)
            assert set(e.composition.elements).issubset(elements)

        assert sorted_entries != entries

    def test_get_phonon_data_by_material_id(self, mpr):
        bs = mpr.get_phonon_bandstructure_by_material_id("mp-661")
        assert isinstance(bs, PhononBandStructureSymmLine)

    def test_get_charge_density_data(self, mpr):
        task_ids = mpr.get_charge_density_calculation_ids_from_material_id("mp-149")
        assert len(task_ids) > 0

        vasp_calc_details = mpr.get_charge_density_calculation_details(
            task_ids[0]["task_id"]
        )
        assert isinstance(vasp_calc_details.incar, Incar)

        chgcar = mpr.get_charge_density_from_calculation_id(task_ids[0]["task_id"])
        assert isinstance(chgcar, Chgcar)

    def test_get_substrates(self, mpr):
        substrate_data = mpr.get_substrates("mp-123", [1, 0, 0])
        substrates = [sub_dict["sub_id"] for sub_dict in substrate_data]
        assert "mp-2534" in substrates

    def test_get_surface_data(self, mpr):
        data = mpr.get_surface_data("mp-126")  # Pt
        one_surf = mpr.get_surface_data("mp-129", miller_index=[-2, -3, 1])
        assert one_surf["surface_energy"] == pytest.approx(2.99156963)
        assert one_surf["miller_index"] == pytest.approx([3, 2, 1])
        assert "surfaces" in data
        surfaces = data["surfaces"]
        assert len(surfaces) > 0
        surface = surfaces.pop()
        assert "miller_index" in surface
        assert "surface_energy" in surface
        assert "is_reconstructed" in surface
        assert "structure" in surface

    @pytest.mark.xfail  # temporary
    def test_get_gb_data(self, mpr):
        mo_gbs = mpr.get_gb_data(chemsys="Mo")
        assert len(mo_gbs) == 10
        mo_gbs_s5 = mpr.get_gb_data(pretty_formula="Mo", sigma=5)
        assert len(mo_gbs_s5) == 3
        mo_s3_112 = mpr.get_gb_data(
            material_id="mp-129", sigma=3, gb_plane=[1, -1, -2],
        )
        assert len(mo_s3_112) == 1
        gb_f = mo_s3_112[0]["final_structure"]
        assert gb_f.rotation_axis == pytest.approx([1, 1, 0])
        assert gb_f.rotation_angle == pytest.approx(109.47122)
        assert mo_s3_112[0]["gb_energy"] == pytest.approx(0.4796547330588574)
        assert mo_s3_112[0]["w_sep"] == pytest.approx(6.318144)

    def test_get_wulff_shape(self, mpr):
        ws = mpr.get_wulff_shape("mp-126")
        assert isinstance(ws, WulffShape)

    def test_query(self, mpr):

        excluded_params = [
            "sort_field",
            "ascending",
            "chunk_size",
            "num_chunks",
            "all_fields",
            "fields",
            "equillibrium_reaction_energy",  # Fix (thermo model)
            "weighted_work_function",  # Fix (search builder)
            "weighted_surface_energy",  # Fix (search builder)
            "surface_anisotropy",  # Fix (search builder)
            "shape_factor",  # Fix (search builder)
        ]

        alt_name_dict = {
            "material_ids": "material_id",
            "chemsys_formula": "formula_pretty",
            "piezoelectric_modulus": "e_ij_max",
            "crystal_system": "symmetry",
            "spacegroup_symbol": "symmetry",
            "spacegroup_number": "symmetry",
            "total_energy": "energy_per_atom",
            "formation_energy": "formation_energy_per_atom",
            "uncorrected_energy": "uncorrected_energy_per_atom",
            "magnetic_ordering": "ordering",
            "elastic_anisotropy": "universal_anisotropy",
            "poisson_ratio": "homogeneous_poisson",
            "piezoelectric_modulus": "e_ij_max",
        }  # type: dict

        custom_field_tests = {
            "material_ids": ["mp-149"],
            "chemsys_formula": "SiO2",
            "crystal_system": CrystalSystem.cubic,
            "spacegroup_number": 38,
            "spacegroup_symbol": "Amm2",
            "magnetic_ordering": Ordering.FM,
            "has_props": ["dielectric"],
            "theoretical": True,
        }  # type: dict

        search_method = mpr.query

        # Get list of parameters
        param_tuples = list(typing.get_type_hints(search_method).items())

        # Query API for each numeric and bollean parameter and check if returned
        for entry in param_tuples:
            param = entry[0]
            if param not in excluded_params:
                param_type = entry[1].__args__[0]
                q = None

                if param in custom_field_tests:
                    project_field = alt_name_dict.get(param, None)
                    q = {
                        param: custom_field_tests[param],
                        "chunk_size": 1,
                        "num_chunks": 1,
                    }
                elif param_type is typing.Tuple[int, int]:
                    project_field = alt_name_dict.get(param, None)
                    q = {
                        param: (-100, 100),
                        "chunk_size": 1,
                        "num_chunks": 1,
                    }
                elif param_type is typing.Tuple[float, float]:
                    project_field = alt_name_dict.get(param, None)
                    q = {
                        param: (-100.12, 100.12),
                        "chunk_size": 1,
                        "num_chunks": 1,
                    }
                elif param_type is bool:
                    project_field = alt_name_dict.get(param, None)
                    q = {
                        param: False,
                        "chunk_size": 1,
                        "num_chunks": 1,
                    }

                doc = search_method(**q)[0].dict()

                assert (
                    doc[project_field if project_field is not None else param]
                    is not None
                )



from enum import Enum
from typing import List

from pydantic import BaseModel, Field, validator
from monty.json import MontyDecoder
from datetime import datetime


class MagnetismData(BaseModel):
    """
    Model for magnetic data within a magnetism doc
    """

    ordering: str = Field(
        None, description="Magnetic ordering.",
    )

    is_magnetic: bool = Field(
        None, description="Whether the material is magnetic.",
    )

    exchange_symmetry: int = Field(
        None, description="Exchange symmetry.",
    )

    num_magnetic_sites: int = Field(
        None, description="The number of magnetic sites.",
    )

    num_unique_magnetic_sites: int = Field(
        None, description="The number of unique magnetic sites.",
    )

    types_of_magnetic_species: List[str] = Field(
        None, description="Magnetic specie elements.",
    )

    magmoms: List[float] = Field(
        None, description="Magnetic moments for each site.",
    )

    total_magnetization: float = Field(
        None, description="Total magnetization in μB.",
    )

    total_magnetization_normalized_vol: float = Field(
        None, description="Total magnetization normalized by volume in μB/Å³.",
    )

    total_magnetization_normalized_formula_units: float = Field(
        None, description="Total magnetization normalized by formula unit in μB/f.u. .",
    )


class MagnetismDoc(BaseModel):
    """
    Magnetic ordering, total magnetizaiton, ...
    """

    task_id: str = Field(
        None,
        description="The ID of this material, used as a universal reference across property documents."
        "This comes in the form: mp-******",
    )

    magnetism: MagnetismData = Field(
        None, description="Magnetic data for the material",
    )

    last_updated: datetime = Field(
        None,
        description="Timestamp for the most recent calculation for this Material document",
    )

    # Make sure that the datetime field is properly formatted
    @validator("last_updated", pre=True)
    def last_updated_dict_ok(cls, v):
        return MontyDecoder().process_decoded(v)



# coding: utf-8
""" Primary MAPI module """
from pkg_resources import get_distribution, DistributionNotFound
from monty.serialization import loadfn
from mp_api.matproj import MPRester

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    __version__ = None  # type: ignore

from mp_api.core.settings import MAPISettings
from pathlib import Path

settings = MAPISettings()

try:  # pragma: no cover
    if Path(settings.app_path).exists():
        mapi = loadfn(settings.app_path)
        app = mapi.app
    else:
        app = None
        if settings.debug:
            print(f"Failed loading App at {settings.app_path}")

except Exception as e:  # pragma: no cover
    # Something went wrong with loading default app
    if settings.debug:
        print("Failed loading App")
        print(e)
        app = None



from typing import List, Optional

from mp_api.routes.tasks.models import TaskDoc
from mp_api.core.client import BaseRester


class TaskRester(BaseRester):

    suffix = "tasks"
    document_model = TaskDoc  # type: ignore
    primary_key = "task_id"

    def get_trajectory(self, task_id):
        """
        Returns a Trajectory object containing the geometry of the
        material throughout a calculation. This is most useful for
        observing how a material relaxes during a geometry optimization.

        :param task_id: A specified task_id
        :return: Trajectory object
        """

        pass

    def search_task_docs(
        self,
        chemsys_formula: Optional[str] = None,
        num_chunks: Optional[int] = None,
        chunk_size: int = 1000,
        all_fields: bool = True,
        fields: Optional[List[str]] = None,
    ):
        """
        Query core task docs using a variety of search criteria.

        Arguments:
            chemsys_formula (str): A chemical system (e.g., Li-Fe-O),
                or formula including anonomyzed formula
                or wild cards (e.g., Fe2O3, ABO3, Si*).
            num_chunks (int): Maximum number of chunks of data to yield. None will yield all possible.
            chunk_size (int): Number of data entries per chunk. Max size is 100.
            all_fields (bool): Whether to return all fields in the document. Defaults to True.
            fields (List[str]): List of fields in TaskDoc to return data for.
                Default is material_id, last_updated, and formula_pretty if all_fields is False.

        Returns:
            ([TaskDoc]) List of task documents
        """

        query_params = {}  # type: dict

        if chemsys_formula:
            query_params.update({"formula": chemsys_formula})

        return super().search(
            num_chunks=num_chunks,
            chunk_size=chunk_size,
            all_fields=all_fields,
            fields=fields,
            **query_params
        )



from maggma.api.resource import ReadOnlyResource
from mp_api.routes.phonon.models import PhononBSDoc, PhononImgDoc
from mp_api.routes.phonon.query_operators import PhononImgQuery

from maggma.api.query_operator import PaginationQuery, SparseFieldsQuery


def phonon_bs_resource(phonon_bs_store):
    resource = ReadOnlyResource(
        phonon_bs_store,
        PhononBSDoc,
        query_operators=[
            PaginationQuery(),
            SparseFieldsQuery(PhononBSDoc, default_fields=["task_id", "last_updated"]),
        ],
        tags=["Phonon"],
        enable_default_search=False,
    )

    return resource


def phonon_img_resource(phonon_img_store):

    resource = ReadOnlyResource(
        phonon_img_store,
        PhononImgDoc,
        tags=["Phonon"],
        enable_default_search=False,
        enable_get_by_key=True,
        key_fields=["plot", "task_id", "last_updated"],
        sub_path="/image/",
    )

    return resource



from typing import List

from pydantic import BaseModel, Field, validator


class ImageEntry(BaseModel):
    """
    Model for an high resolution image entry in SurfacePropDoc
    """

    miller_index: List[int] = Field(
        None, description="Miller index of surface.",
    )

    image: bytes = Field(
        None, description="High resolution image data.",
    )

    # Make sure that the datetime field is properly formatted
    @validator("image", pre=True)
    def image_bytes_ok(cls, v):
        return str(v)


class WulffDoc(BaseModel):
    """
    Model for a document containing wulff construction image data.
    """

    hi_res_images: List[ImageEntry] = Field(
        None, description="List of individual surface image data.",
    )

    pretty_formula: str = Field(
        None, description="Reduced formula of the material.",
    )

    thumbnail: str = Field(
        None, description="Thumbnail image data.",
    )

    task_id: str = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******",
    )

    # Make sure that the datetime field is properly formatted
    @validator("thumbnail", pre=True)
    def thumbnail_bytes_ok(cls, v):
        return str(v)



from platform import version
from mp_api.core.client import BaseRester, MPRestError
from mp_api.routes.tasks.client import TaskRester
from mp_api.routes.materials.client import MaterialsRester


class ChargeDensityRester(BaseRester):

    suffix = "charge_density"
    primary_key = "task_id"

    def get_charge_density_from_calculation_id(self, task_id: str):
        """
        Get charge density data for a given calculation ID.

        Arguments:
            task_id (str): Calculation ID

        Returns:
            chgcar (dict): Pymatgen CHGCAR object.
        """

        result = self.search(task_ids=task_id, fields=["task_id", "data"], chunk_size=1)

        if len(result) > 0:
            return result[0]["data"]
        else:
            raise MPRestError("No charge density found")

    def get_charge_density_calculation_details(self, task_id: str):
        """
        Get charge density calculations details for a given calculation (task) ID.

        Arguments:
            task_id (str): Calculation (task) ID

        Returns:
            calc_details (dict): Dictionary containing INCAR, POSCAR, and KPOINTS data for the DFT calculation.
        """

        task_rester = TaskRester(endpoint=self.base_endpoint, api_key=self.api_key)  # type: ignore

        result = task_rester.get_document_by_id(
            document_id=task_id,
            fields=["orig_inputs.incar", "orig_inputs.poscar", "orig_inputs.kpoints"],
        ).orig_inputs

        return result

    def get_charge_density_calculation_ids_from_material_id(self, material_id: str):
        """
        Get charge density calculation ids associated with a given Materials Project ID
        that have charge density data.

        Arguments:
            material_id (str): Materials Project ID

        Returns:
            calculation_ids (List[str]): List of calculation ids that have charge density data.
        """

        materials_rester = MaterialsRester(  # type: ignore
            endpoint=self.base_endpoint, api_key=self.api_key,
        )

        mat_doc = materials_rester.get_document_by_id(
            document_id=material_id, fields=["calc_types"]
        )

        calculation_ids = []
        if mat_doc is not None:
            for calculation_id, calculation_type in mat_doc.calc_types.items():
                if "Static" in calculation_type:
                    calculation_ids.append(calculation_id)

        result = []

        if len(calculation_ids) > 0:
            result = self.search(
                task_ids=",".join(calculation_ids), fields=["task_id"], chunk_size=10
            )

        return result



from mp_api.routes.phonon.query_operators import PhononImgQuery

from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn


def test_phonon_image_query():
    op = PhononImgQuery()

    assert op.query(task_id="mp-149") == {"criteria": {"task_id": "mp-149"}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(task_id="mp-149") == {"criteria": {"task_id": "mp-149"}}



from emmet.core.search import SearchDoc

from maggma.api.query_operator import (
    PaginationQuery,
    SortQuery,
    SparseFieldsQuery,
    NumericQuery,
)
from maggma.api.resource import ReadOnlyResource, AggregationResource
from mp_api.routes.materials.query_operators import (
    DeprecationQuery,
    ElementsQuery,
    FormulaQuery,
    SymmetryQuery,
)
from mp_api.routes.search.models import SearchStats
from mp_api.routes.search.query_operators import (
    HasPropsQuery,
    MaterialIDsSearchQuery,
    SearchIsStableQuery,
    SearchIsTheoreticalQuery,
    SearchMagneticQuery,
    SearchStatsQuery,
    SearchESQuery,
)


def search_resource(search_store):
    resource = ReadOnlyResource(
        search_store,
        SearchDoc,
        query_operators=[
            MaterialIDsSearchQuery(),
            FormulaQuery(),
            ElementsQuery(),
            SymmetryQuery(),
            SearchIsStableQuery(),
            SearchIsTheoreticalQuery(),
            SearchMagneticQuery(),
            SearchESQuery(),
            NumericQuery(model=SearchDoc, excluded_fields=["composition"]),
            HasPropsQuery(),
            DeprecationQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(SearchDoc, default_fields=["material_id"]),
        ],
        tags=["Search"],
    )

    return resource


def search_stats_resource(search_store):
    resource = AggregationResource(
        search_store,
        SearchStats,
        pipeline_query_operator=SearchStatsQuery(SearchDoc),
        tags=["Search"],
        sub_path="/stats/",
    )

    return resource



from typing import List, Optional
from mp_api.routes.fermi.models import FermiDoc


class FermiRester:

    def get_document_by_id(
        self,
        document_id: str,
        fields: Optional[List[str]] = None,
        monty_decode: bool = True,
    ) -> FermiDoc: ...



from typing import List, Optional
from emmet.core.electrode import InsertionElectrodeDoc


class ElectrodeRester:

    def query_by_task_id(
        self, task_id, fields: Optional[List[str]] = None, monty_decode: bool = True
    ) -> InsertionElectrodeDoc: ...



from maggma.api.resource import ReadOnlyResource
from mp_api.routes.magnetism.models import MagnetismDoc

from maggma.api.query_operator import PaginationQuery, SortQuery, SparseFieldsQuery
from mp_api.routes.magnetism.query_operators import MagneticQuery


def magnetism_resource(magnetism_store):
    resource = ReadOnlyResource(
        magnetism_store,
        MagnetismDoc,
        query_operators=[
            MagneticQuery(),
            SortQuery(),
            PaginationQuery(),
            SparseFieldsQuery(MagnetismDoc, default_fields=["task_id", "last_updated"]),
        ],
        tags=["Magnetism"],
    )

    return resource



from typing import Optional, Tuple, List
from collections import defaultdict
from mp_api.core.client import BaseRester, MPRestError
from mp_api.routes.electronic_structure.models.core import BSPathType, DOSProjectionType
from emmet.core.electronic_structure import ElectronicStructureDoc
from pymatgen.analysis.magnetism.analyzer import Ordering
from pymatgen.electronic_structure.core import Spin, OrbitalType
from pymatgen.core.periodic_table import Element


class ElectronicStructureRester(BaseRester):

    suffix = "electronic_structure"
    document_model = ElectronicStructureDoc  # type: ignore
    primary_key = "material_id"

    def search_electronic_structure_docs(
        self,
        band_gap: Optional[Tuple[float, float]] = None,
        efermi: Optional[Tuple[float, float]] = None,
        magnetic_ordering: Optional[Ordering] = None,
        is_gap_direct: bool = None,
        is_metal: bool = None,
        sort_field: Optional[str] = None,
        ascending: Optional[bool] = None,
        num_chunks: Optional[int] = None,
        chunk_size: int = 1000,
        all_fields: bool = True,
        fields: Optional[List[str]] = None,
    ):
        """
        Query electronic structure docs using a variety of search criteria.

        Arguments:
            band_gap (Tuple[float,float]): Minimum and maximum band gap in eV to consider.
            efermi (Tuple[float,float]): Minimum and maximum fermi energy in eV to consider.
            magnetic_ordering (Ordering): Magnetic ordering of the material.
            is_gap_direct (bool): Whether the material has a direct band gap.
            is_metal (bool): Whether the material is considered a metal.
            sort_field (str): Field used to sort results.
            ascending (bool): Whether sorting should be in ascending order.
            num_chunks (int): Maximum number of chunks of data to yield. None will yield all possible.
            chunk_size (int): Number of data entries per chunk.
            all_fields (bool): Whether to return all fields in the document. Defaults to True.
            fields (List[str]): List of fields in EOSDoc to return data for.
                Default is material_id and last_updated if all_fields is False.

        Returns:
            ([ElectronicStructureDoc]) List of electronic structure documents
        """

        query_params = defaultdict(dict)  # type: dict

        if band_gap:
            query_params.update(
                {"band_gap_min": band_gap[0], "band_gap_max": band_gap[1]}
            )

        if efermi:
            query_params.update({"efermi_min": efermi[0], "efermi_max": efermi[1]})

        if magnetic_ordering:
            query_params.update({"magnetic_ordering": magnetic_ordering.value})

        if is_gap_direct is not None:
            query_params.update({"is_gap_direct": is_gap_direct})

        if is_metal is not None:
            query_params.update({"is_metal": is_metal})

        if sort_field:
            query_params.update({"sort_field": sort_field})

        if ascending is not None:
            query_params.update({"ascending": ascending})

        query_params = {
            entry: query_params[entry]
            for entry in query_params
            if query_params[entry] is not None
        }

        return super().search(
            num_chunks=num_chunks,
            chunk_size=chunk_size,
            all_fields=all_fields,
            fields=fields,
            **query_params
        )


class BandStructureRester(BaseRester):

    suffix = "electronic_structure/bandstructure"
    document_model = ElectronicStructureDoc  # type: ignore

    def search_bandstructure_summary(
        self,
        path_type: BSPathType = BSPathType.setyawan_curtarolo,
        band_gap: Optional[Tuple[float, float]] = None,
        efermi: Optional[Tuple[float, float]] = None,
        magnetic_ordering: Optional[Ordering] = None,
        is_gap_direct: bool = None,
        is_metal: bool = None,
        sort_field: Optional[str] = None,
        ascending: Optional[bool] = None,
        num_chunks: Optional[int] = None,
        chunk_size: int = 1000,
        all_fields: bool = True,
        fields: Optional[List[str]] = None,
    ):
        """
        Query band structure summary data in electronic structure docs using a variety of search criteria.

        Arguments:
            path_type (BSPathType): k-path selection convention for the band structure.
            band_gap (Tuple[float,float]): Minimum and maximum band gap in eV to consider.
            efermi (Tuple[float,float]): Minimum and maximum fermi energy in eV to consider.
            magnetic_ordering (Ordering): Magnetic ordering of the material.
            is_gap_direct (bool): Whether the material has a direct band gap.
            is_metal (bool): Whether the material is considered a metal.
            sort_field (str): Field used to sort results.
            ascending (bool): Whether sorting should be in ascending order.
            num_chunks (int): Maximum number of chunks of data to yield. None will yield all possible.
            chunk_size (int): Number of data entries per chunk.
            all_fields (bool): Whether to return all fields in the document. Defaults to True.
            fields (List[str]): List of fields in ElectronicStructureDoc to return data for.
                Default is material_id and last_updated if all_fields is False.

        Returns:
            ([ElectronicStructureDoc]) List of electronic structure documents
        """

        query_params = defaultdict(dict)  # type: dict

        query_params["path_type"] = path_type.value

        if band_gap:
            query_params.update(
                {"band_gap_min": band_gap[0], "band_gap_max": band_gap[1]}
            )

        if efermi:
            query_params.update({"efermi_min": efermi[0], "efermi_max": efermi[1]})

        if magnetic_ordering:
            query_params.update({"magnetic_ordering": magnetic_ordering.value})

        if is_gap_direct is not None:
            query_params.update({"is_gap_direct": is_gap_direct})

        if is_metal is not None:
            query_params.update({"is_metal": is_metal})

        if sort_field:
            query_params.update({"sort_field": sort_field})

        if ascending is not None:
            query_params.update({"ascending": ascending})

        query_params = {
            entry: query_params[entry]
            for entry in query_params
            if query_params[entry] is not None
        }

        return super().search(
            num_chunks=num_chunks,
            chunk_size=chunk_size,
            all_fields=all_fields,
            fields=fields,
            **query_params
        )

    def get_bandstructure_from_calculation_id(self, task_id: str):
        """
        Get the band structure pymatgen object associated with a given calculation ID.

        Arguments:
            task_id (str): Calculation ID for the band structure calculation

        Returns:
            bandstructure (BandStructure): BandStructure or BandStructureSymmLine object
        """

        result = self._query_resource(
            criteria={"task_id": task_id, "all_fields": True},
            suburl="object",
            use_document_model=False,
        )

        if result.get("data", None) is not None:
            return result["data"]
        else:
            raise MPRestError("No object found")

    def get_bandstructure_from_material_id(
        self,
        material_id: str,
        path_type: BSPathType = BSPathType.setyawan_curtarolo,
        line_mode=True,
    ):
        """
        Get the band structure pymatgen object associated with a Materials Project ID.

        Arguments:
            materials_id (str): Materials Project ID for a material
            path_type (BSPathType): k-point path selection convention
            line_mode (bool): Whether to return data for a line-mode calculation

        Returns:
            bandstructure (Union[BandStructure, BandStructureSymmLine]): BandStructure or BandStructureSymmLine object
        """

        es_rester = ElectronicStructureRester(
            endpoint=self.base_endpoint, api_key=self.api_key
        )

        if line_mode:
            bs_data = es_rester.get_document_by_id(
                document_id=material_id, fields=["bandstructure"]
            ).bandstructure.dict()

            if bs_data.get(path_type.value, None):
                bs_calc_id = bs_data[path_type.value]["task_id"]
            else:
                raise MPRestError(
                    "No {} band structure data found for {}".format(
                        path_type.value, material_id
                    )
                )
        else:
            bs_data = es_rester.get_document_by_id(
                document_id=material_id, fields=["dos"]
            ).dos.dict()

            if bs_data.get("total", None):
                bs_calc_id = bs_data["total"]["1"]["task_id"]
            else:
                raise MPRestError(
                    "No uniform band structure data found for {}".format(material_id)
                )

        bs_obj = self.get_bandstructure_from_calculation_id(bs_calc_id)

        if bs_obj:
            return bs_obj[0]["data"]
        else:
            raise MPRestError("No band structure object found.")


class DosRester(BaseRester):

    suffix = "electronic_structure/dos"
    document_model = ElectronicStructureDoc  # type: ignore

    def search_dos_summary(
        self,
        projection_type: DOSProjectionType = DOSProjectionType.total,
        spin: Spin = Spin.up,
        element: Optional[Element] = None,
        orbital: Optional[OrbitalType] = None,
        band_gap: Optional[Tuple[float, float]] = None,
        efermi: Optional[Tuple[float, float]] = None,
        magnetic_ordering: Optional[Ordering] = None,
        num_chunks: Optional[int] = None,
        chunk_size: int = 1000,
        all_fields: bool = True,
        fields: Optional[List[str]] = None,
    ):
        """
        Query density of states summary data in electronic structure docs using a variety of search criteria.

        Arguments:
            projection_type (DOSProjectionType): Projection type of dos data. Default is the total dos.
            spin (Spin): Spin channel of dos data. If non spin-polarized data is stored in Spin.up
            element (Element): Element for element-projected dos data.
            orbital (OrbitalType): Orbital for orbital-projected dos data.
            band_gap (Tuple[float,float]): Minimum and maximum band gap in eV to consider.
            efermi (Tuple[float,float]): Minimum and maximum fermi energy in eV to consider.
            magnetic_ordering (Ordering): Magnetic ordering of the material.
            num_chunks (int): Maximum number of chunks of data to yield. None will yield all possible.
            chunk_size (int): Number of data entries per chunk.
            all_fields (bool): Whether to return all fields in the document. Defaults to True.
            fields (List[str]): List of fields in EOSDoc to return data for.
                Default is material_id and last_updated if all_fields is False.

        Returns:
            ([ElectronicStructureDoc]) List of electronic structure documents
        """

        query_params = defaultdict(dict)  # type: dict

        query_params["projection_type"] = projection_type.value
        query_params["spin"] = spin.value

        if element:
            query_params["element"] = element.value

        if orbital:
            query_params["orbital"] = orbital.value

        if band_gap:
            query_params.update(
                {"band_gap_min": band_gap[0], "band_gap_max": band_gap[1]}
            )

        if efermi:
            query_params.update({"efermi_min": efermi[0], "efermi_max": efermi[1]})

        if magnetic_ordering:
            query_params.update({"magnetic_ordering": magnetic_ordering.value})

        query_params = {
            entry: query_params[entry]
            for entry in query_params
            if query_params[entry] is not None
        }

        return super().search(
            num_chunks=num_chunks,
            chunk_size=chunk_size,
            all_fields=all_fields,
            fields=fields,
            **query_params
        )

    def get_dos_from_calculation_id(self, task_id: str):
        """
        Get the density of states pymatgen object associated with a given calculation ID.

        Arguments:
            task_id (str): Calculation ID for the density of states calculation

        Returns:
            bandstructure (CompleteDos): CompleteDos object
        """

        result = self._query_resource(
            criteria={"task_id": task_id, "all_fields": True},
            suburl="object",
            use_document_model=False,
        )

        if result.get("data", None) is not None:
            return result["data"]
        else:
            raise MPRestError("No object found")

    def get_dos_from_material_id(self, material_id: str):
        """
        Get the complete density of states pymatgen object associated with a Materials Project ID.

        Arguments:
            materials_id (str): Materials Project ID for a material

        Returns:
            dos (CompleteDos): CompleteDos object
        """

        es_rester = ElectronicStructureRester(
            endpoint=self.base_endpoint, api_key=self.api_key
        )

        dos_data = es_rester.get_document_by_id(
            document_id=material_id, fields=["dos"]
        ).dict()

        if dos_data["dos"]:
            dos_calc_id = dos_data["dos"]["total"]["1"]["task_id"]
        else:
            raise MPRestError(
                "No density of states data found for {}".format(material_id)
            )

        dos_obj = self.get_dos_from_calculation_id(dos_calc_id)
        if dos_obj:
            return dos_obj[0]["data"]
        else:
            raise MPRestError("No density of states object found.")
